#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <SDL/SDL.h>
#include "mini3d.h"

#if defined(__unix__) || defined(unix)
#include <unistd.h>
#else	/* win32 */
#include <windows.h>
#define usleep(x) Sleep(x)
#endif	/* unix / win32 */

int init(void);
void update_gfx(void);
void clean_up(void);
void draw_cube(float sz);
void parse_cmdline(int argc, char **argv);
void *load_image(const char *fname, unsigned long *xsz, unsigned long *ysz);

int xsz = 640;
int ysz = 480;
const char *img_fname;
unsigned int tex;
SDL_Surface *fbsurf;
int model = 0;

int auto_rot = 1;
int phong_shading = 0;
float xrot, yrot, zrot;

#define MESH_NVERT	5535
#define MESH_NNORM	29124
#define MESH_NFACE	9708

extern float vertices[MESH_NVERT][3];
extern float normals[MESH_NNORM][3];
extern int triangles[MESH_NFACE][3];


int main(int argc, char **argv) {
	parse_cmdline(argc, argv);

	if(init() == -1) {
		return EXIT_FAILURE;
	}
	
	for(;;) {
		SDL_Event event;
		if(SDL_PollEvent(&event)) {
			if(event.type == SDL_KEYDOWN) {
				if(event.key.keysym.sym == 27) {
					break;
				}

				switch(event.key.keysym.sym) {
				case 32:
					model = (model + 1) % 2;
					break;

				case 'r':
					auto_rot = !auto_rot;
					break;

				case 'p':
					phong_shading = !phong_shading;
					if(phong_shading) {
						m3d_enable(M3D_PHONG);
					} else {
						m3d_disable(M3D_PHONG);
					}

				case SDLK_LEFT:
					yrot -= 5;
					break;

				case SDLK_RIGHT:
					yrot += 5;
					break;

				case SDLK_UP:
					xrot -= 5;
					break;

				case SDLK_DOWN:
					xrot += 5;
					break;

				default:
					break;
				}
			}
		} else {
			/* some supporting code to make this sleep when it doesn't need the CPU, ignore */
			const static unsigned int frame_interval = 33;	/* 30 fps (1000 / 30) */
			static int prev_time;
			int frame_time, cur_time;
			
			update_gfx();
			
			frame_time = (cur_time = SDL_GetTicks()) - prev_time;
			if(frame_time < frame_interval) {
				usleep((frame_interval - frame_time) * 1000);
			}

			prev_time = cur_time;
		}
	}

	clean_up();
	return 0;
}


int init(void) {
	float lpos[] = {-100, 100, 100, 1.0};
	SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER);
	if(!(fbsurf = SDL_SetVideoMode(xsz, ysz, 32, SDL_SWSURFACE))) {
		fputs("SDL init failed\n", stderr);
		return -1;
	}

	SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY / 2, SDL_DEFAULT_REPEAT_INTERVAL);

	/* initialize mini3d */
	m3d_init();
	m3d_viewport(0, 0, xsz, ysz);
	
	/* set the color/depth that m3d_clear() uses to clear the buffers */
	m3d_clear_color(0, 0, 0, 0);
	m3d_clear_depth(1);

	/* enable depth-testing (zbuffering) and lighting */
	m3d_enable(M3D_DEPTH_TEST);
	m3d_enable(M3D_LIGHTING);
	m3d_enable(M3D_LIGHT0);
	/*m3d_enable(M3D_PHONG);*/

	m3d_light(M3D_LIGHT0, M3D_POSITION, lpos);

	m3d_light_model(M3D_LIGHT_MODEL_LOCAL_VIEWER, 1);
	/*
	m3d_enable(M3D_BLEND);
	m3d_blend_func(M3D_SRC_ALPHA, M3D_ONE_MINUS_SRC_ALPHA);
	*/

	if(img_fname) {
		uint32_t *img;
		unsigned long tx, ty;
		
		if(!(img = load_image(img_fname, &tx, &ty))) {
			fprintf(stderr, "failed to load %s\n", img_fname);
			return -1;
		}

		m3d_enable(M3D_TEXTURE_2D);
		m3d_gen_textures(1, &tex);
		m3d_bind_texture(M3D_TEXTURE_2D, tex);
		m3d_tex_image(M3D_TEXTURE_2D, 0, 4, tx, ty, 0, M3D_BGRA, M3D_UNSIGNED_BYTE, img);
		free(img);
	}

	/* setup a projection matrix */
	m3d_matrix_mode(M3D_PROJECTION);
	m3d_load_identity();
	m3d_perspective(45.0, 1.33333, 100.0, 1000.0);

	/* prepare the material for rendering the object */
	{
		float dif[] = {0.2, 0.4, 1.0, 1.0};
		float spec[] = {1, 1, 1, 1};
		m3d_materialv(M3D_DIFFUSE, dif);
		m3d_materialv(M3D_SPECULAR, spec);
		m3d_material(M3D_SHININESS, 60.0f);
	}

	return 0;
}

void update_gfx(void) {
	int i, j;
	uint32_t *pixels;
	float t = (float)SDL_GetTicks() / 20.0f;

	/* clear the buffers */
	m3d_clear(M3D_COLOR_BUFFER_BIT | M3D_DEPTH_BUFFER_BIT);

	/* set the current modelview transformation matrix */
	m3d_matrix_mode(M3D_MODELVIEW);
	m3d_load_identity();
	m3d_translate(0, 0, -500);

	if(auto_rot) {
		if(model == 1) m3d_rotate(t, 1, 0, 0);
		m3d_rotate(t, 0, 1, 0);
	} else {
		m3d_rotate_euler(xrot, yrot, zrot);
	}

	/* render the triangles of the object using m3d_vertex/m3d_normal calls
	 * to specify the positions and normal vectors for each vertex
	 */
	if(model == 0) {
		m3d_begin(M3D_TRIANGLES);
		for(i=0; i<MESH_NFACE; i++) {
			for(j=0; j<3; j++) {
				int nidx = i * 3 + j;
				int vidx = triangles[i][j];
			
				m3d_normal(normals[nidx][0], normals[nidx][1], normals[nidx][2]);
				m3d_vertex(vertices[vidx][0], vertices[vidx][1], vertices[vidx][2]);
			}
		}
		m3d_end();
	} else {
		draw_cube(200);
	}
	
	pixels = m3d_get_pixel_data();	/* get a pointer to the framebuffer ... */

	/* ... and copy it to the SDL surface */
	SDL_LockSurface(fbsurf);
	memcpy(fbsurf->pixels, pixels, xsz * ysz * 4);
	SDL_UnlockSurface(fbsurf);
	
	SDL_Flip(fbsurf);	/* finally show the surface */
}

void clean_up(void) {
	m3d_destroy();
	SDL_Quit();
}

void draw_cube(float sz) {
	sz *= 0.5;
	m3d_begin(M3D_QUADS);
	
	/* face +Z */
	m3d_normal(0, 0, 1);
	m3d_color3(1, 0, 0);
	m3d_tex_coord(0, 0);
	m3d_vertex(sz, -sz, sz);
	m3d_tex_coord(1, 0);
	m3d_vertex(-sz, -sz, sz);
	m3d_tex_coord(1, 1);
	m3d_vertex(-sz, sz, sz);
	m3d_tex_coord(0, 1);
	m3d_vertex(sz, sz, sz);

	/* face -X */
	m3d_normal(-1, 0, 0);
	m3d_color3(1, 0, 0);
	m3d_tex_coord(0, 0);
	m3d_vertex(-sz, -sz, sz);
	m3d_tex_coord(1, 0);
	m3d_vertex(-sz, -sz, -sz);
	m3d_tex_coord(1, 1);
	m3d_vertex(-sz, sz, -sz);
	m3d_tex_coord(0, 1);
	m3d_vertex(-sz, sz, sz);

	/* face -Z */
	m3d_normal(0, 0, -1);
	m3d_color3(1, 0, 0);
	m3d_tex_coord(0, 1);
	m3d_vertex(-sz, sz, -sz);
	m3d_tex_coord(0, 0);
	m3d_vertex(-sz, -sz, -sz);
	m3d_tex_coord(1, 0);
	m3d_vertex(sz, -sz, -sz);
	m3d_tex_coord(1, 1);
	m3d_vertex(sz, sz, -sz);

	/* face +X */
	m3d_normal(1, 0, 0);
	m3d_color3(1, 0, 0);
	m3d_tex_coord(0, 1);
	m3d_vertex(sz, sz, -sz);
	m3d_tex_coord(0, 0);
	m3d_vertex(sz, -sz, -sz);
	m3d_tex_coord(1, 0);
	m3d_vertex(sz, -sz, sz);
	m3d_tex_coord(1, 1);
	m3d_vertex(sz, sz, sz);

	/* face +Y */
	m3d_normal(0, 1, 0);
	m3d_color3(1, 0, 0);
	m3d_tex_coord(1, 0);
	m3d_vertex(sz, sz, sz);
	m3d_tex_coord(0, 0);
	m3d_vertex(-sz, sz, sz);
	m3d_tex_coord(0, 1);
	m3d_vertex(-sz, sz, -sz);
	m3d_tex_coord(1, 1);
	m3d_vertex(sz, sz, -sz);

	/* face -Y */
	m3d_normal(0, -1, 0);
	m3d_color3(1, 0, 0);
	m3d_tex_coord(1, 0);
	m3d_vertex(-sz, -sz, -sz);
	m3d_tex_coord(1, 1);
	m3d_vertex(-sz, -sz, sz);
	m3d_tex_coord(0, 1);
	m3d_vertex(sz, -sz, sz);
	m3d_tex_coord(0, 0);
	m3d_vertex(sz, -sz, -sz);

	m3d_end();
}

void parse_cmdline(int argc, char **argv) {
	int i;
	char *sep;

	for(i=1; i<argc; i++) {
		if(argv[i][0] == '-' && argv[i][2] == 0) {
			switch(argv[i][1]) {
			case 's':
				if(!isdigit(argv[++i][0]) || !(sep = strchr(argv[i], 'x')) || !isdigit(*(sep+1))) {
					fprintf(stderr, "malformed -s argument: \"%s\"\n", argv[i]);
					exit(EXIT_FAILURE);
				}
				xsz = atoi(argv[i]);
				ysz = atoi(sep + 1);
				break;

			case 't':
				img_fname = argv[++i];
				break;

			case 'h':
			default:
				fprintf(stderr, "usage: example [-s WxH] [-t tex_fname]\n");
				exit(argv[i][1] == 'h' ? 0 : EXIT_FAILURE);
			}
		} else {
			fprintf(stderr, "unrecognized argument: %s\n", argv[i]);
			exit(EXIT_FAILURE);
		}
	}
}
